@@ -26,7 +26,7 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with this program. If not, see <http://www.gnu.org/licenses/>.
+# along with this program. If not, see <https://www.gnu.org/licenses/>.
 
 use 5.008;  # We're using PerlIO layers
 use strict;
@@ -35,12 +35,14 @@ use open ':utf8';  # changelogs are written with UTF-8 encoding
 use filetest 'access';  # use access rather than stat for -w
 # for checking whether user names are valid and making format() behave
 use Encode qw/decode_utf8 encode_utf8/;
-use Getopt::Long qw(:config gnu_getopt);
+use Getopt::Long qw(:config bundling permute no_getopt_compat);
 use File::Copy;
 use File::Basename;
 use Cwd;
 use Dpkg::Vendor qw(get_current_vendor);
-use lib '/usr/share/devscripts';
+use Dpkg::Changelog::Parse qw(changelog_parse);
+use Dpkg::Control;
+use Dpkg::Version;
 use Devscripts::Compression;
 use Devscripts::Debbugs;
 use POSIX qw(locale_h strftime);
@@ -59,25 +61,6 @@ my $CHGLINE;  # used by the format O section at the end
 
 my $compression_re = compression_get_file_extension_regex();
 
-my $lpdc_broken;
-
-sub have_lpdc {
-    return ($lpdc_broken ? 0 : 1) if defined $lpdc_broken;
-    eval {
-	require Parse::DebControl;
-    };
-
-    if ($@) {
-	if ($@ =~ m%^Can\'t locate Parse/DebControl%) {
-	    $lpdc_broken="the libparse-debcontrol-perl package is not installed";
-	} else {
-	    $lpdc_broken="couldn't load Parse::DebControl: $@";
-	}
-    }
-    else { $lpdc_broken=''; }
-    return $lpdc_broken ? 0 : 1;
-}
-
 my $debian_distro_info;
 sub get_debian_distro_info {
     return $debian_distro_info if defined $debian_distro_info;
@@ -178,7 +161,9 @@ Options:
          distribution name
   --bpo
          Increment the Debian release number for a backports upload
-         to "wheezy-backports"
+         to "stretch-backports"
+  --stable
+         Increment the Debian release number for a stable upload.
   -l, --local <suffix>
          Add a suffix to the Debian version number for a local build
   -b, --force-bad-version
@@ -246,7 +231,7 @@ Options:
   --version
          Display version information
   At most one of -a, -i, -e, -r, -v, -d, -n, --bin-nmu, -q, --qa, -R, -s,
-  --team, --bpo, -l (or their long equivalents) may be used.
+  --team, --bpo, --stable, -l (or their long equivalents) may be used.
   With no options, one of -i or -a is chosen by looking at the release
   specified in the changelog.
 
@@ -257,7 +242,7 @@ EOF
 
 sub version () {
     print <<"EOF";
-This is $progname, from the Debian devscripts package, version ###VERSION###
+This is $progname, from the Debian devscripts package, version 2.17.10
 This code is copyright 1999-2003 by Julian Gilbey, all rights reserved.
 Based on code by Christoph Lameter.
 This program comes with ABSOLUTELY NO WARRANTY.
@@ -366,7 +351,7 @@ if (@ARGV and $ARGV[0] =~ /^--no-?conf$/) {
 # with older debchange versions.
 my ($opt_help, $opt_version);
 my ($opt_i, $opt_a, $opt_e, $opt_r, $opt_v, $opt_b, $opt_d, $opt_D, $opt_u, $opt_force_dist);
-my ($opt_n, $opt_bn, $opt_qa, $opt_R, $opt_s, $opt_team, $opt_U, $opt_bpo, $opt_l, $opt_c, $opt_m, $opt_M, $opt_create, $opt_package, @closes);
+my ($opt_n, $opt_bn, $opt_qa, $opt_R, $opt_s, $opt_team, $opt_U, $opt_bpo, $opt_stable, $opt_l, $opt_c, $opt_m, $opt_M, $opt_create, $opt_package, @closes);
 my ($opt_news);
 my ($opt_level, $opt_regex, $opt_noconf, $opt_empty);
 
@@ -396,6 +381,7 @@ GetOptions("help|h" => \$opt_help,
 	   "team" => \$opt_team,
 	   "U|upstream" => \$opt_U,
 	   "bpo" => \$opt_bpo,
+	   "stable" => \$opt_stable,
 	   "l|local=s" => \$opt_l,
 	   "query!" => \$opt_query,
 	   "closes=s" => \@closes,
@@ -447,8 +433,8 @@ if (defined $opt_level) {
 if (defined $opt_regex) { $check_dirname_regex = $opt_regex; }
 
 # Only allow at most one non-help option
-fatal "Only one of -a, -i, -e, -r, -v, -d, -n/--nmu, --bin-nmu, -q/--qa, -R/--rebuild, -s/--security, --team, --bpo, -l/--local is allowed;\ntry $progname --help for more help"
-    if ($opt_i?1:0) + ($opt_a?1:0) + ($opt_e?1:0) + ($opt_r?1:0) + ($opt_v?1:0) + ($opt_d?1:0) + ($opt_n?1:0) + ($opt_bn?1:0) + ($opt_qa?1:0) + ($opt_R?1:0) + ($opt_s?1:0) + ($opt_team?1:0) + ($opt_bpo?1:0) + ($opt_l?1:0) > 1;
+fatal "Only one of -a, -i, -e, -r, -v, -d, -n/--nmu, --bin-nmu, -q/--qa, -R/--rebuild, -s/--security, --team, --bpo, --stable, -l/--local is allowed;\ntry $progname --help for more help"
+    if ($opt_i?1:0) + ($opt_a?1:0) + ($opt_e?1:0) + ($opt_r?1:0) + ($opt_v?1:0) + ($opt_d?1:0) + ($opt_n?1:0) + ($opt_bn?1:0) + ($opt_qa?1:0) + ($opt_R?1:0) + ($opt_s?1:0) + ($opt_team?1:0) + ($opt_bpo?1:0) + ($opt_stable?1:0) + ($opt_l?1:0) > 1;
 
 if ($opt_s) {
     $opt_u = "high";
@@ -461,7 +447,7 @@ if (defined $opt_u) {
 
 # See if we're Debian, Ubuntu or someone else, if we can
 my $vendor;
-if (not $opt_vendor eq '') {
+if (defined $opt_vendor && $opt_vendor) {
     $vendor = $opt_vendor;
 } else {
     if (defined $opt_D) {
@@ -482,14 +468,14 @@ if (not $opt_vendor eq '') {
     }
 }
 $vendor ||= 'Debian';
-if ($vendor eq 'Ubuntu' and ($opt_n or $opt_bn or $opt_qa or $opt_bpo)) {
+if ($vendor eq 'Ubuntu' and ($opt_n or $opt_bn or $opt_qa or $opt_bpo or $opt_stable)) {
     $vendor = 'Debian';
 }
 
 # Check the distro name given.
 if (defined $opt_D) {
     if ($vendor eq 'Debian') {
-	unless ($opt_D =~ /^(experimental|unstable|UNRELEASED|((old)?stable|testing)(-proposed-updates|-security)?|proposed-updates)$/) {
+	unless ($opt_D =~ /^(experimental|unstable|sid|UNRELEASED|((old){0,2}stable|testing|wheezy|jessie|stretch|buster|bullseye)(-proposed-updates|-security)?|proposed-updates)$/) {
 	    my $deb_info = get_debian_distro_info();
 	    my ($oldstable_backports, $stable_backports) = ("", "");
 	    if ($deb_info == 0) {
@@ -504,9 +490,12 @@ if (defined $opt_D) {
 	    if ($deb_info == 0 || $opt_D !~ m/^(\Q$stable_backports\E|\Q$oldstable_backports\E)$/) {
 		$stable_backports = ", " . $stable_backports if $stable_backports;
 		$oldstable_backports = ", " . $oldstable_backports if $oldstable_backports;
-		warn "$progname warning: Recognised distributions are: unstable, testing, stable,\n"
-		     . "oldstable, experimental, {testing-,stable-,oldstable-,}proposed-updates,\n"
-		     . "{testing,stable,oldstable}-security$oldstable_backports$stable_backports and UNRELEASED.\n"
+		warn "$progname warning: Recognised distributions are: \n"
+		     . "experimental, unstable, testing, stable, oldstable, oldoldstable,\n"
+                     . "{bullseye,buster,stretch,jessie,wheezy}-proposed-updates,\n"
+		     . "{testing,stable,oldstable,oldoldstable}-proposed-updates,\n"
+                     . "{bullseye,buster,stretch,jessie,wheezy}-security,\n"
+		     . "{testing,stable,oldstable,oldoldstable}}-security$oldstable_backports$stable_backports and UNRELEASED.\n"
 		     . "Using your request anyway.\n";
 		$warnings++ if not $opt_force_dist;
 	    }
@@ -554,13 +543,13 @@ fatal "--package cannot be used when creating a NEWS file"
 
 if ($opt_create) {
     if ($opt_a || $opt_i || $opt_e || $opt_r || $opt_b || $opt_n || $opt_bn ||
-	    $opt_qa || $opt_R || $opt_s || $opt_team || $opt_bpo || $opt_l ||
-	    $opt_allow_lower) {
-	warn "$progname warning: ignoring -a/-i/-e/-r/-b/--allow-lower-version/-n/--bin-nmu/-q/--qa/-R/-s/--team/--bpo/-l options with --create\n";
-	$warnings++;
+            $opt_qa || $opt_R || $opt_s || $opt_team || $opt_bpo ||
+            $opt_stable || $opt_l || $opt_allow_lower) {
+        warn "$progname warning: ignoring -a/-i/-e/-r/-b/--allow-lower-version/-n/--bin-nmu/-q/--qa/-R/-s/--team/--bpo/--stable,-l options with --create\n";
+        $warnings++;
     }
     if ($opt_package && $opt_d) {
-	fatal "Can only use one of --package and -d";
+        fatal "Can only use one of --package and -d";
     }
 }
 
@@ -616,93 +605,59 @@ else {  # $opt_create
 #####
 
 # Find the current version number etc.
-my %changelog;
+my $changelog;
 my $PACKAGE = 'PACKAGE';
 my $VERSION = 'VERSION';
 my $MAINTAINER = 'MAINTAINER';
 my $EMAIL = 'EMAIL';
 my $DISTRIBUTION = 'UNRELEASED';
 my $bpo_dist = '';
-my %bpo_dists = ( 60, 'squeeze', 70, 'wheezy' );
-my $latest_bpo_dist = '70';
+my %bpo_dists = ( 60, 'squeeze', 70, 'wheezy', 8, 'jessie', 9, 'stretch' );
+my $stable_dist = '';
+my %stable_dists = ( 8, 'jessie', 9, 'stretch' );
+my $latest_dist = '9';
 my $CHANGES = '';
 # Changelog urgency, possibly propogated to NEWS files
 my $CL_URGENCY = '';
 
 if (! $opt_create || ($opt_create && $opt_news)) {
-    if (! $opt_create) {
-	open PARSED, qq[dpkg-parsechangelog -l"$changelog_path" | ]
-	    or fatal "Cannot execute dpkg-parsechangelog: $!";
-    } elsif ($opt_create && $opt_news) {
-	open PARSED, qq[dpkg-parsechangelog | ]
-	    or fatal "Cannot execute dpkg-parsechangelog: $!";
-    } else {
-	fatal "This can't happen: what am I parsing?";
-    }
-
-    my $last;
-    while (<PARSED>) {
-	chomp;
-	if (/^(\S+):\s(.+?)\s*$/) { $changelog{$1}=$2; $last=$1; }
-	elsif (/^(\S+):\s$/) { $changelog{$1}=''; $last=$1; }
-	elsif (/^\s\.$/) { $changelog{$last}.="\n"; }
-	elsif (/^\s(.+)$/) { $changelog{$last}.="$1\n"; }
-	else {
-	    fatal "Don't understand dpkg-parsechangelog output: $_";
-	}
-    }
-
-    close PARSED
-	or fatal "Problem executing dpkg-parsechangelog: $!";
-    if ($?) { fatal "dpkg-parsechangelog failed!"; }
+    my $file = $opt_create ? 'debian/changelog' : $changelog_path;
+    $changelog = changelog_parse(file => $file);
 
     # Now we've read the changelog, set some variables and then
     # let's check the directory name is sensible
     fatal "No package name in changelog!"
-	unless exists $changelog{'Source'};
-    $PACKAGE = $changelog{'Source'};
+	unless exists $changelog->{Source};
+    $PACKAGE = $changelog->{Source};
     fatal "No version number in changelog!"
-	unless exists $changelog{'Version'};
-    $VERSION=$changelog{'Version'};
+	unless exists $changelog->{Version};
+    $VERSION=$changelog->{Version};
     fatal "No maintainer in changelog!"
-	unless exists $changelog{'Maintainer'};
-    ($MAINTAINER,$EMAIL) = ($changelog{'Maintainer'} =~ /^([^<]+) <(.*)>/);
+	unless exists $changelog->{Maintainer};
+    $changelog->{Maintainer} = decode_utf8($changelog->{Maintainer});
+    ($MAINTAINER,$EMAIL) = ($changelog->{Maintainer} =~ /^([^<]*) <(.*)>/);
+    $MAINTAINER ||= '';
     fatal "No distribution in changelog!"
-	unless exists $changelog{'Distribution'};
+	unless exists $changelog->{Distribution};
     if ($vendor eq 'Ubuntu') {
 	# In Ubuntu the development release regularly changes, don't just copy
 	# the previous name.
 	$DISTRIBUTION=get_ubuntu_devel_distro();
     } else {
-	$DISTRIBUTION=$changelog{'Distribution'};
+	$DISTRIBUTION=$changelog->{Distribution};
     }
     fatal "No changes in changelog!"
-	unless exists $changelog{'Changes'};
+	unless exists $changelog->{Changes};
 
     # Find the current package version
     if ($opt_news) {
 	my $found_version = 0;
 	my $found_urgency = 0;
-	open PARSED, qq[dpkg-parsechangelog -l"$real_changelog_path" | ]
-	    or fatal "Cannot execute dpkg-parsechangelog: $!";
-	while (<PARSED>) {
-	    chomp;
-	    if (m%^Version:\s+(\S+)$%) {
-		$VERSION = $1;
-		$VERSION =~ s/~$//;
-		$found_version = 1;
-		last if $found_urgency;
-	    } elsif (m%^Urgency:\s+(\S+)(\s|$)%) {
-		$CL_URGENCY = $1;
-		$found_urgency = 1;
-		last if $found_version;
-	    } elsif (m%^$%) {
-		last;
-	    }
-	}
-	close PARSED
-	    or fatal "Problem executing dpkg-parsechangelog: $!";
-	if ($?) { fatal "dpkg-parsechangelog failed!"; }
+	my $clog = changelog_parse(file => $real_changelog_path);
+	$VERSION = $clog->{Version};
+	$VERSION =~ s/~$//;
+
+	$CL_URGENCY = $clog->{Urgency};
     }
 
     # Is the directory name acceptable?
@@ -849,18 +804,14 @@ if (! $opt_m and ! $opt_M) {
 
 if ($opt_M) {
     if (-f 'debian/control') {
-	if (have_lpdc()) {
-	    my $parser = Parse::DebControl->new;
-	    my $deb822 = $parser->parse_file('debian/control', {stripComments => 'true'});
-	    my $maintainer = decode_utf8($deb822->[0]->{'Maintainer'});
-	    if ($maintainer =~ /^(.*)\s+<(.*)>$/) {
-		$MAINTAINER = $1;
-		$EMAIL = $2;
-	    } else {
-		fatal "$progname: invalid debian/control Maintainer field value\n";
-	    }
+	my $parser = Dpkg::Control->new(type => CTRL_INFO_SRC);
+	$parser->load('debian/control');
+	my $maintainer = decode_utf8($parser->{Maintainer});
+	if ($maintainer =~ /^(.*)\s+<(.*)>$/) {
+	    $MAINTAINER = $1;
+	    $EMAIL = $2;
 	} else {
-	    fatal "$progname: unable to get maintainer from debian/control: $lpdc_broken\n";
+	    fatal "$progname: invalid debian/control Maintainer field value\n";
 	}
     } else {
 	fatal "Missing file debian/control";
@@ -871,31 +822,27 @@ if ($opt_M) {
 
 if ($opt_auto_nmu eq 'yes' and ! $opt_v and ! $opt_l and ! $opt_s and
     ! $opt_team and ! $opt_qa and ! $opt_R and ! $opt_bpo and ! $opt_bn and
-    ! $opt_n and ! $opt_c and
+    ! $opt_n and ! $opt_c and ! $opt_stable and
     ! (exists $ENV{'CHANGELOG'} and length $ENV{'CHANGELOG'}) and ! $opt_M and
     ! $opt_create and ! $opt_a_passed and ! $opt_r and ! $opt_e and
     $vendor ne 'Ubuntu' and $vendor ne 'Tanglu' and
     ! ($opt_release_heuristic eq 'changelog' and
-       $changelog{'Distribution'} eq 'UNRELEASED' and ! $opt_i_passed)) {
+       $changelog->{Distribution} eq 'UNRELEASED' and ! $opt_i_passed)) {
 
     if (-f 'debian/control') {
-	if (have_lpdc()) {
-	    my $parser = new Parse::DebControl;
-	    my $deb822 = $parser->parse_file('debian/control', {stripComments => 'true'});
-	    my $uploader = decode_utf8($deb822->[0]->{'Uploaders'}) || '';
-	    my $maintainer = decode_utf8($deb822->[0]->{'Maintainer'});
-	    my @uploaders = split(/,\s*/, $uploader);
+	my $parser = Dpkg::Control->new(type => CTRL_INFO_SRC);
+	$parser->load('debian/control');
+	my $uploader = decode_utf8($parser->{Uploaders}) || '';
+	my $maintainer = decode_utf8($parser->{Maintainer});
+	my @uploaders = split(/,\s*/, $uploader);
 
-	    my $packager = "$MAINTAINER <$EMAIL>";
+	my $packager = "$MAINTAINER <$EMAIL>";
 
-	    if ($maintainer !~ m/<packages\@qa\.debian\.org>/ and
-		! grep { $_ eq $packager } ($maintainer, @uploaders) and
-		$packager ne $changelog{'Maintainer'} and ! $opt_team) {
-		$opt_n=1;
-		$opt_a=0;
-	    }
-	} else {
-	    warn "$progname: skipping automatic NMU detection: $lpdc_broken\n";
+	if ($maintainer !~ m/<packages\@qa\.debian\.org>/ and
+	    ! grep { $_ eq $packager } ($maintainer, @uploaders) and
+	    $packager ne $changelog->{Maintainer} and ! $opt_team) {
+	    $opt_n=1;
+	    $opt_a=0;
 	}
     } else {
 	fatal "Missing file debian/control";
@@ -1001,7 +948,7 @@ my $DATE;
 }
 
 if ($opt_news && !$opt_i && !$opt_a) {
-    if ($VERSION eq $changelog{'Version'} && !$opt_v && !$opt_l) {
+    if ($VERSION eq $changelog->{Version} && !$opt_v && !$opt_l) {
 	$opt_a = 1;
     } else {
 	$opt_i = 1;
@@ -1011,7 +958,7 @@ if ($opt_news && !$opt_i && !$opt_a) {
 # Are we going to have to figure things out for ourselves?
 if (! $opt_i && ! $opt_v && ! $opt_d && ! $opt_a && ! $opt_e && ! $opt_r &&
     ! $opt_n && ! $opt_bn && ! $opt_qa && ! $opt_R && ! $opt_s && ! $opt_team &&
-    ! $opt_bpo && ! $opt_l && ! $opt_create) {
+    ! $opt_bpo && ! $opt_stable && ! $opt_l && ! $opt_create) {
     # Yes, we are
     if ($opt_release_heuristic eq 'log') {
 	my @UPFILES = glob("../$PACKAGE\_$SVERSION\_*.upload");
@@ -1041,7 +988,7 @@ if (! $opt_i && ! $opt_v && ! $opt_d && ! $opt_a && ! $opt_e && ! $opt_r &&
 	}
     }
     elsif ($opt_release_heuristic eq 'changelog') {
-	if ($changelog{'Distribution'} eq 'UNRELEASED') {
+	if ($changelog->{Distribution} eq 'UNRELEASED') {
 		$opt_a = 1;
 	}
 	elsif ($EMPTY_TEXT==1) {
@@ -1060,7 +1007,7 @@ unless ($opt_create) {
     open S, $changelog_path or fatal "Cannot open existing $changelog_path: $!";
 
     # Read the first stanza from the changelog file
-    # We do this directly rather than reusing $changelog{'Changes'}
+    # We do this directly rather than reusing $changelog->{Changes}
     # so that we have the verbatim changes rather than a (albeit very
     # slightly) reformatted version. See Debian bug #452806
 
@@ -1088,8 +1035,8 @@ my $optionsok=0;
 my $merge=0;
 
 if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
-     $opt_bpo || $opt_l || $opt_v || $opt_d ||
-    ($opt_news && $VERSION ne $changelog{'Version'})) && ! $opt_create) {
+     $opt_bpo || $opt_stable || $opt_l || $opt_v || $opt_d ||
+    ($opt_news && $VERSION ne $changelog->{Version})) && ! $opt_create) {
 
     $optionsok=1;
 
@@ -1132,8 +1079,7 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
 	    $opt_p=1;
 	}
 
-	if (system("dpkg --compare-versions $VERSION le $NEW_VERSION" .
-		  " 2>/dev/null 1>&2")) {
+	if (version_compare($VERSION, $NEW_VERSION) == 1) {
 	    if ($opt_b or ($opt_allow_lower and $NEW_VERSION =~ /$opt_allow_lower/)) {
 		warn "$progname warning: new version ($NEW_VERSION) is less than\n" .
 		    "the current version number ($VERSION).\n";
@@ -1185,12 +1131,14 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
 	             not $start =~ /build/ and not $start =~ /ubuntu/) {
 		$end .= "build1";
 	    } elsif ($opt_R and $vendor eq 'Tanglu' and
-	             not $start =~ /b/ and not $start =~ /tanglu/) {
+	             not "$start$end" =~ /(b\d+)$/ and not $start =~ /tanglu/) {
 		$end .= "b1";
 	    } elsif ($opt_bpo and not $start =~ /~bpo[0-9]+\+$/) {
 		# If it's not already a backport make it so
 		# otherwise we can be safe if we behave like dch -i
-		$end .= "~bpo$latest_bpo_dist+1";
+		$end .= "~bpo$latest_dist+1";
+	    } elsif ($opt_stable and not $start =~ /\+deb\d+u/) {
+		$end .= "+deb${latest_dist}u1";
 	    } elsif ($opt_l and not $start =~ /\Q$opt_l\E/) {
 		# If it's not already a local package make it so
 		# otherwise we can be safe if we behave like dch -i
@@ -1210,10 +1158,10 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
 		} elsif (($opt_i or $opt_s) and $vendor eq 'Tanglu' and
 		     $start !~ /(tanglu)(\d+\.)*$/ and not $opt_U) {
 
-		    if ($start =~ /b/) {
-			# Drop bX suffix in favor of tanglu1
-			$start =~ s/b//;
-			$end = "";
+		    if ("$start$end" =~ /(b\d+)$/) {
+				# Drop bX suffix in favor of tanglu1
+				$start =~ s/b$//;
+				$end = "";
 		    }
 		    $end .= "tanglu1";
 		} else {
@@ -1224,16 +1172,31 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
 		# based on the version of the previous backport
 		if ($opt_bpo) {
 		    my $previous_dist = $start;
-		    $previous_dist =~ s/^.*~bpo([0-9]+)\+$/$1/;
+		    $previous_dist =~ s/^.*~bpo(\d+)\+$/$1/;
 		    if (defined $previous_dist and defined
 			$bpo_dists{$previous_dist}) {
 			$bpo_dist = $bpo_dists{$previous_dist} . '-backports';
 		    } else {
 			# Fallback to using the previous distribution
-			$bpo_dist = $changelog{'Distribution'};
+			$bpo_dist = $changelog->{Distribution};
 		    }
 		}
 
+		# Attempt to set the distribution for a stable upload correctly
+		# based on the version of the previous upload
+		if ($opt_stable) {
+		    my $previous_dist = $start;
+		    $previous_dist =~ s/^.*+deb(\d+)u$/$1/;
+		    if (defined $previous_dist and defined
+			$stable_dists{$previous_dist}) {
+			$stable_dist = $stable_dists{$previous_dist};
+		    } else {
+			# Fallback to using the previous distribution
+			$stable_dist = $changelog->{Distribution};
+		    }
+		}
+
+
 		if(! ($opt_s or $opt_n or $vendor eq 'Ubuntu' or $vendor eq 'Tanglu')) {
 		    if ($start =~/(.*?)-(\d+)\.$/) {
 			# Drop NMU revision
@@ -1245,7 +1208,7 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
 		    }
 		}
 
-		if (! ($opt_qa or $opt_bpo or $opt_l)) {
+		if (! ($opt_qa or $opt_bpo or $opt_stable or $opt_l)) {
 		    $useextra = 1;
 		}
 	    }
@@ -1268,9 +1231,12 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
     }
 
     if ($opt_bpo) {
-	$bpo_dist ||= $bpo_dists{$latest_bpo_dist} . '-backports';
+	$bpo_dist ||= $bpo_dists{$latest_dist} . '-backports';
     }
-    my $distribution = $opt_D || $bpo_dist || (($opt_release_heuristic eq 'changelog') ? "UNRELEASED" : $DISTRIBUTION);
+    if ($opt_stable) {
+	$stable_dist ||= $stable_dists{$latest_dist};
+    }
+    my $distribution = $opt_D || $bpo_dist || $stable_dist || (($opt_release_heuristic eq 'changelog') ? "UNRELEASED" : $DISTRIBUTION);
 
     my $urgency = $opt_u;
     if ($opt_news) {
@@ -1279,13 +1245,14 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
     $urgency ||= 'medium';
 
     if (($opt_v or $opt_i or $opt_l or $opt_d) and
-	$opt_release_heuristic eq "changelog" and
-	$changelog{'Distribution'} eq "UNRELEASED" and
-	$distribution eq "UNRELEASED") {
+	$opt_release_heuristic eq 'changelog' and
+	$changelog->{Distribution} eq 'UNRELEASED') {
 
 	$merge = 1;
     } else {
-	print O "$PACKAGE ($NEW_VERSION) $distribution; urgency=$urgency\n\n";
+	print O "$PACKAGE ($NEW_VERSION) $distribution; urgency=$urgency";
+	print O ", binary-only=yes" if ($opt_bn);
+	print O "\n\n";
 	if ($opt_n && ! $opt_news) {
 	    print O "  * Non-maintainer upload.\n";
 	    $line = 1;
@@ -1311,7 +1278,7 @@ if (($opt_i || $opt_n || $opt_bn || $opt_qa || $opt_R || $opt_s || $opt_team ||
 	    print O "  * Rebuild for $bpo_dist.\n";
 	    $line = 1;
 	}
-	if (@closes_text or $TEXT) {
+	if (@closes_text or $TEXT or $EMPTY_TEXT) {
 	    foreach (@closes_text) { format_line($_, 1); }
 	    if (length $TEXT) { format_line($TEXT, 1); }
 	} elsif ($opt_news) {
@@ -1370,13 +1337,16 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
 		if ($tmpver =~ m/~bpo(\d+)\+/ && exists $bpo_dists{$1}) {
 		    $dist_indicator = "$bpo_dists{$1}-backports";
 		}
+		if ($tmpver =~ m/\+deb(\d+)u/ && exists $stable_dists{$1}) {
+		    $dist_indicator = "$stable_dists{$1}";
+		}
 	    }
 	}
 	elsif (/  \* (?:Upload to|Rebuild for) (\S+).*$/) {
 	    ($dist_indicator = $1) =~ s/[!:.,;]$//;
 	    chomp $dist_indicator;
 	}
-	elsif (/^ --\s+([^<]+)\s+/) {
+	elsif (/^ --\s+([^<]+)\s+/ || /^ --\s+<(.+?)>/) {
 	    $lastmaint=$1;
 	    # Remember where we are so we can skip back afterwards
 	    $savedline = $line;
@@ -1469,7 +1439,8 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
 
     if (! $opt_r) {
 	# Add a multi-maintainer header...
-	if ($multimaint) {
+	if ($multimaint and
+		(@closes_text or $TEXT or $opt_news or !$EMPTY_TEXT)) {
 	    # ...unless there already is one for this maintainer.
 	    if (!defined $maintline) {
 		print O "\n  [ $MAINTAINER ]\n";
@@ -1499,7 +1470,7 @@ if (($opt_r || $opt_a || $merge) && ! $opt_create) {
     }
 
     if ($opt_t && $opt_a) {
-	print O "\n -- $changelog{'Maintainer'}  $changelog{'Date'}\n";
+	print O "\n -- $changelog->{Maintainer}  $changelog->{Date}\n";
     } else {
 	print O "\n -- $MAINTAINER <$EMAIL>  $DATE\n";
     }
@@ -1520,7 +1491,7 @@ elsif ($opt_e && ! $opt_create) {
     print O $CHANGES;
 
     if ($opt_t) {
-	print O "\n -- $changelog{'Maintainer'}  $changelog{'Date'}\n";
+	print O "\n -- $changelog->{Maintainer}  $changelog->{Date}\n";
     } else {
 	print O "\n -- $MAINTAINER <$EMAIL>  $DATE\n";
     }
@@ -1614,24 +1585,18 @@ copy("$changelog_path.dch","$changelog_path") or
 if ((basename(cwd()) =~ m%^\Q$PACKAGE\E-\Q$UVERSION\E$%) &&
     !$opt_p && !$opt_create) {
     # Find the current version number etc.
-    my ($new_version, $new_sversion, $new_uversion);
-    open PARSED, "dpkg-parsechangelog |"
-	or fatal "Cannot execute dpkg-parsechangelog: $!";
-    while (<PARSED>) {
-	if (/^Version:\s(.+?)\s*$/) { $new_version=$1; }
+    my $v;
+    my $changelog = changelog_parse();
+    if (exists $changelog->{Version}) {
+	$v = Dpkg::Version->new($changelog->{Version});
     }
 
-    close PARSED
-	or fatal "Problem executing dpkg-parsechangelog: $!";
-    if ($?) { fatal "dpkg-parsechangelog failed!" }
-
     fatal "No version number in debian/changelog!"
-	unless defined $new_version;
+	unless $v;
 
-    # Is this a native Debian package, i.e., does it have a - in the
-    # version number?
-    $new_version =~ s/^\d+://;  # remove epoch
-    ($new_uversion=$new_version) =~ s/-[^-]*$//;  # remove revision
+    my ($new_version, $new_uversion);
+    $new_version = $v->as_string(omit_epoch => 1);
+    $new_uversion = $v->as_string(omit_epoch => 1, omit_revision => 1);
 
     if ($new_uversion ne $UVERSION) {
 	# Then we rename the directory
@@ -1640,11 +1605,13 @@ if ((basename(cwd()) =~ m%^\Q$PACKAGE\E-\Q$UVERSION\E$%) &&
 	} else {
 	    warn "$progname warning: Couldn't rename directory: $!\n";
 	}
-	# And check whether a new orig tarball exists
-	my @origs = glob("../$PACKAGE\_$new_uversion.*");
-	my $num_origs = grep { /^..\/\Q$PACKAGE\E_\Q$new_uversion\E\.orig\.tar\.$compression_re$/ } @origs;
-	if ($num_origs == 0) {
-	    warn "$progname warning: no orig tarball found for the new version.\n";
+	if (!$v->is_native()) {
+	    # And check whether a new orig tarball exists
+	    my @origs = glob("../$PACKAGE\_$new_uversion.*");
+	    my $num_origs = grep { /^..\/\Q$PACKAGE\E_\Q$new_uversion\E\.orig\.tar\.$compression_re$/ } @origs;
+	    if ($num_origs == 0) {
+		warn "$progname warning: no orig tarball found for the new version.\n";
+	    }
 	}
     }
 }
@@ -1676,7 +1643,7 @@ sub format_line {
 
     # Work around the fact that write() with formats
     # seems to assume that characters are single-byte
-    # See http://rt.perl.org/rt3/Public/Bug/Display.html?id=33832
+    # See https://rt.perl.org/Public/Bug/Display.html?id=33832
     # and Debian bugs #473769 and #541484
     # This relies on $CHGLINE being a sequence of unicode characters.  We can
     # compare how many unicode characters we have to how many bytes we have
